/* src/net/ocsp_response.c
 * ============================================================
 * OCSP Responder
 * (c) 2001 by Massimiliano Pala
 * OpenCA released software
 * ============================================================
 */
 
#include "general.h"

#include <openssl/err.h>
#include <openssl/ssl.h>

#include <openssl/txt_db.h>
// #include <openssl/asn1.h>

#include "support.h"
#include "db.h"

#include "ocsp_response.h"


/* External General Variables */
extern int debug;
extern int verbose;

int make_ocsp_response(OCSP_RESPONSE **resp, OCSP_REQUEST *req,
		OCSPD_CONFIG *conf )
	{
	ASN1_TIME *thisupd = NULL, *nextupd = NULL;
	OCSP_CERTID *cid, *ca_id = NULL;
	OCSP_BASICRESP *bs = NULL;
	int i, id_count, ret = 1;


	id_count = OCSP_request_onereq_count(req);

	if (id_count <= 0)
		{
		*resp = OCSP_response_create(OCSP_RESPONSE_STATUS_MALFORMEDREQUEST, NULL);
		goto end;
		}

	ca_id = OCSP_cert_to_id(conf->digest, NULL, conf->cacert);

	bs = OCSP_BASICRESP_new();
	thisupd = X509_gmtime_adj(NULL, 0);
	if (conf->ndays != -1)
		nextupd = X509_gmtime_adj(NULL, conf->nmin * 60 + 
				conf->ndays * 3600 * 24 );

	/* Examine each certificate id in the request */
	for (i = 0; i < id_count; i++)
		{
		OCSP_ONEREQ *one;
		ASN1_INTEGER *serial;
		char inf;

		one = OCSP_request_onereq_get0(req, i);
		cid = OCSP_onereq_get0_id(one);

		/* Get basic request info */
		OCSP_id_get0_info(NULL, NULL, NULL, &serial, cid);

		/* Is this request about our CA? */
		if (OCSP_id_issuer_cmp(ca_id, cid))
			{
			if (verbose)
				{
				syslog( LOG_INFO, "request for non reckognized CA %ld",
					ASN1_INTEGER_get(serial) );
				}

			OCSP_basic_add1_status(bs, cid,
						V_OCSP_CERTSTATUS_UNKNOWN,
						0, NULL,
						thisupd, nextupd);
			continue;
			}

		if( verbose )
			{
			syslog( LOG_INFO, "request for certificate serial %ld",
					ASN1_INTEGER_get( serial ) );
			}

		inf = db_get_status(serial, conf);
		if (inf == DB_TYPE_UNK)
			{
			if( verbose )
				{
				syslog( LOG_INFO, "status unknown for %ld certificate",
					ASN1_INTEGER_get(serial) );
				}
			
			OCSP_basic_add1_status(bs, cid,
						V_OCSP_CERTSTATUS_UNKNOWN,
						0, NULL,
						thisupd, nextupd);
			}
		else if (inf == DB_TYPE_VAL) 
		 	{
			if( verbose )
				{
				syslog( LOG_INFO, "status VALID for %ld certificate",
					ASN1_INTEGER_get(serial) );
				}
			
			OCSP_basic_add1_status(bs, cid,
						V_OCSP_CERTSTATUS_GOOD,
						0, NULL,
						thisupd, nextupd);
		 	}
		else 
			{
			ASN1_OBJECT *inst = NULL;
			ASN1_TIME *revtm = NULL;
			ASN1_GENERALIZEDTIME *invtm = NULL;
			OCSP_SINGLERESP *single;

			int reason = -1;
			unpack_revinfo(&revtm, &reason, &inst, 
					&invtm, db_get_revdate(serial, conf));

			single = OCSP_basic_add1_status(bs, cid,
						V_OCSP_CERTSTATUS_REVOKED,
						reason, revtm,
						thisupd, nextupd);
			if (invtm)
				OCSP_SINGLERESP_add1_ext_i2d(single, 
					NID_invalidity_date, invtm, 0, 0);
			else if (inst)
				OCSP_SINGLERESP_add1_ext_i2d(single, 
					NID_hold_instruction_code, inst, 0, 0);

			if( verbose )
				{
				if(invtm)
					syslog( LOG_INFO, "status REVOKED for %ld (inval. date)",
						ASN1_INTEGER_get(serial) );
				else if (inst)
					syslog( LOG_INFO, "status REVOKED for %ld (hold inst. code)",
						ASN1_INTEGER_get(serial) );
				else
					syslog( LOG_INFO, "status REVOKED for %ld (none)",
						ASN1_INTEGER_get(serial) );
				}

			ASN1_OBJECT_free(inst);
			ASN1_TIME_free(revtm);
			ASN1_GENERALIZEDTIME_free(invtm);
			}
		}

	OCSP_copy_nonce(bs, req);
	OCSP_basic_sign(bs, conf->ocspd_cert, conf->ocspd_pkey, 
			conf->digest, conf->other_certs,
			conf->flags);

	*resp = OCSP_response_create(OCSP_RESPONSE_STATUS_SUCCESSFUL, bs);

	end:
	ASN1_TIME_free(thisupd);
	ASN1_TIME_free(nextupd);
	OCSP_CERTID_free(ca_id);
	OCSP_BASICRESP_free(bs);
	return ret;

}
