/*
 * getcertinitial.c -- handle a GetCertInitial message
 *
 * (c) 2001 Dr. Andreas Mueller, Beratung und Entwicklung
 *
 * $Id: getcertinitial.c,v 1.3 2001/03/11 14:58:54 afm Exp $
 */
#include <stdlib.h>
#include <stdio.h>
#include <scep.h>
#include <init.h>
#include <getcertinitial.h>
#include <goodreply.h>
#include <badreply.h>
#include <config.h>
#include <sys/types.h>
#include <sys/stat.h>

/*
 * there are the following possible reactions to a GetCertInitial
 * request
 * - if the certificate has been granted, it can be returned to the
 *   caller
 * - if the certificate is still pending, a pending reply is sent to
 *   the caller
 * - if the certificate was rejected, a failure reply is sent to the
 *   caller
 */
int	getcertinitial(scep_t *scep) {
	char		filename[1024];
	struct stat	sb;
	scepmsg_t	*msg;
	int		havereq = 0;

	if (debug)
		BIO_printf(bio_err, "%s:%d: handling GetCertInitial\n",
			__FILE__, __LINE__);
	msg = &scep->request;

	/* check whether the certificate request for this transaction	*/
	/* has already been rejected					*/
	snprintf(filename, sizeof(filename), "%s/rejected/%s.der", OPENSCEPDIR,
		scep->transId);
	if (debug)
		BIO_printf(bio_err, "%s:%d: checking rejected request at %s\n",
			__FILE__, __LINE__, filename);
	if (stat(filename, &sb) == 0) {
		if (debug)
			BIO_printf(bio_err, "%s:%d: FAILURE reply to "
				"GetCertInitial\n", __FILE__, __LINE__);
		/* request rejected					*/
		msg->failinfo = SCEP_FAILURE_BADCERTID;
		badreply(scep, SCEP_PKISTATUS_FAILURE);
		return 0;
	}

	/* find out whether the certificate request is still around, so	*/
	/* that we can retrieve the challengePassword from it if we	*/
	/* want to set the password in the LDAP directory based on the	*/
	/* challenge, the request can be in one of two places: either	*/
	/* the cert was already signed, then it is in the granted 	*/
	/* directory, or it wasn't, then the request is in the pending	*/
	/* directory							*/
	scep->clientreq = NULL;
	snprintf(filename, sizeof(filename), "%s/granted/%s.req", OPENSCEPDIR,
		scep->transId);
	if (stat(filename, &sb) == 0) {
		if (debug)
			BIO_printf(bio_err, "%s:%d: request was already "
				"granted\n, retrieving backup\n", __FILE__,
				__LINE__);
		havereq = 1;
	}
	snprintf(filename, sizeof(filename), "%s/pending/%s.der", OPENSCEPDIR,
		scep->transId);
	if (stat(filename, &sb) == 0) {
		if (debug)
			BIO_printf(bio_err, "%s:%d: request pending, reading "
				"request\n", __FILE__, __LINE__);
		havereq = 1;
	}
	if (havereq) {
		BIO	*reqbio;
		reqbio = BIO_new(BIO_s_file());
		BIO_read_filename(reqbio, filename);
		scep->clientreq = d2i_X509_REQ_bio(reqbio, NULL);
		if ((scep->clientreq != NULL) && (debug))
			BIO_printf(bio_err, "%s:%d: got X.509 request\n",
				__FILE__, __LINE__);
	}

	/* if the certificate has been granted, return the certificate	*/
	snprintf(filename, sizeof(filename), "%s/granted/%s.der", OPENSCEPDIR,
		scep->transId);
	if (debug)
		BIO_printf(bio_err, "%s:%d: checking granted request at %s\n",
			__FILE__, __LINE__, filename);
	if (stat(filename, &sb) == 0) {
		if (debug)
			BIO_printf(bio_err, "%s:%d: request granted, preparing "
				"CertRep\n", __FILE__, __LINE__);
		/* request has been granted, prepare a certificate 	*/
		/* reply						*/
		goodreply(scep, 1);
		return 0;
	}
	
	/* check whether the certificate request for this transaction 	*/
	/* id is still pending						*/
	snprintf(filename, sizeof(filename), "%s/pending/%s.der", OPENSCEPDIR,
		scep->transId);
	if (debug)
		BIO_printf(bio_err, "%s:%d: checking pending request at %s\n",
			__FILE__, __LINE__, filename);
	if (stat(filename, &sb) == 0) {
		if (debug)
			BIO_printf(bio_err, "%s:%d: PENDING reply to "
				"GetCertInitial, request %s pending\n",
				__FILE__, __LINE__, filename);
		/* request still pending				*/
		badreply(scep, SCEP_PKISTATUS_PENDING);
		return 0;
	}

	/* it may be the case that the certificate is only known to	*/
	/* the directory, in which case we should retrieve it based	*/
	/* on issuer and subject					*/

	/* XXX missing XXX missing XXX missing				*/

	BIO_printf(bio_err, "%s:%d: don't really know what to do\n",
		__FILE__, __LINE__);
	return -11;
}
