/*
 * pkcsreq.c -- handle a PKCS request
 *
 * (c) 2001 Dr. Andreas Mueller, Beratung und Entwicklung
 *
 * $Id: pkcsreq.c,v 1.8 2001/03/26 10:36:41 afm Exp $
 */
#include <config.h>
#include <init.h>
#include <pkcsreq.h>
#include <openssl/x509.h>
#include <openssl/bio.h>
#include <fingerprint.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <goodreply.h>
#include <badreply.h>
#include <grant.h>
#include <check.h>

int	pkcsreq(scep_t *scep) {
	BIO		*outbio;
	int		i;
	char		filename[1024];	/* XXX should do something more	*/
					/* generic here			*/
	char		subject[1024];
	struct stat	sb;
	scepmsg_t	*msg;
	
	/* pkcsreq messages are created by the server			*/
	msg = &scep->request;

	/* debug message to inform debugger what we are doing		*/
	if (debug)
		BIO_printf(bio_err, "%s:%d: handling PKCSreq message\n",
			__FILE__, __LINE__);

	/* compute the fingerprint of the data found in content/length	*/
	/* fields of the scep structure					*/
	scep->fingerprint = fingerprint(msg->data, msg->length);
	if (debug)
		BIO_printf(bio_err, "%s:%d: the request fingerprint is '%s'\n",
			__FILE__, __LINE__, scep->fingerprint);

	/* extract the public key from the X509_REQ, and fingerprint it	*/
	scep->keyfingerprt = x509_key_fingerprint(msg->rd.req);
	if (debug)
		BIO_printf(bio_err, "%s:%d: key fingerprint is %s\n", 
			__FILE__, __LINE__, scep->keyfingerprt);

	/* make sure the key fingerprint matches the transaction id	*/
	if (fingerprint_cmp(scep->keyfingerprt, scep->transId)) {
		BIO_printf(bio_err, "%s:%d: key fingerprint != transId\n",
			__FILE__, __LINE__);
		syslog(LOG_ERR, "%s:%d: fingerprint does not match transid",
			__FILE__, __LINE__);
		goto err;
	}
	if (debug)
		BIO_printf(bio_err, "%s:%d: key fingerprint matches transId "
			"%s\n", __FILE__, __LINE__, scep->transId);

	/* check whether a certificate for the same key finerprint	*/
	/* has already been granted					*/
	snprintf(filename, sizeof(filename), "%s/granted/%s.der", OPENSCEPDIR,
		scep->transId);
	if (stat(filename, &sb) == 0) {
		BIO_printf(bio_err, "%s:%d: request for this transID already "
			"granted\n", __FILE__, __LINE__);
		syslog(LOG_INFO, "%s:%d: request for this transid already "
			"granted", __FILE__, __LINE__);
		return goodreply(scep, 1);
	}

	/* check for a request with the same transaction ID in the 	*/
	/* queue							*/
	snprintf(filename, sizeof(filename), "%s/pending/%s.der", OPENSCEPDIR,
		scep->transId);
	if (stat(filename, &sb) == 0) {
		BIO_printf(bio_err, "%s:%d: already a request with same id: "
			"'%s'\n", __FILE__, __LINE__, scep->transId);
		syslog(LOG_INFO, "%s:%d: PKCSReq for pending certificate",
			__FILE__, __LINE__);
		/* give a pending reply					*/
		return badreply(scep, SCEP_PKISTATUS_PENDING);
	}

	/* create a file containing the request description		*/
	snprintf(filename, sizeof(filename), "%s/%s/%s.info", OPENSCEPDIR,
		"pending", scep->transId);
	outbio = BIO_new(BIO_s_file());
	BIO_write_filename(outbio, filename);
	X509_NAME_oneline(X509_REQ_get_subject_name(msg->rd.req), subject,
		sizeof(subject));
	BIO_printf(outbio, "subject: %s\n", subject);
	BIO_printf(outbio, "transId: %s\n", scep->transId);
	BIO_printf(outbio, "senderNonce: ");
	for (i = 0; i < scep->senderNonceLength; i++) {
		BIO_printf(outbio, "%02x", scep->senderNonce[i]);
	}
	BIO_printf(outbio, "\n");
	BIO_printf(outbio, "fingerprint: %s\n", scep->fingerprint);
	BIO_printf(outbio, "keyfingerprint: %s\n", scep->keyfingerprt);
	BIO_free(outbio);

	/* remember the request through the pointer clientreq in the	*/
	/* SCEP structure						*/
	scep->clientreq = scep->request.rd.req;
	if (debug)
		BIO_printf(bio_err, "%s:%d: client request is at %p\n",
			__FILE__, __LINE__, scep->clientreq);

	/* write the request DER to a file with name <transId>.der	*/
	outbio = BIO_new(BIO_s_file());
	snprintf(filename, sizeof(filename), "%s/%s/%s.der", OPENSCEPDIR,
		"pending", scep->transId);
	BIO_write_filename(outbio, filename);
	if (i2d_X509_REQ_bio(outbio, msg->rd.req) <= 0) {
		BIO_printf(bio_err, "%s:%d: failed to write request to queue\n",
			__FILE__, __LINE__);
		syslog(LOG_ERR, "%s:%d: failed to write request to queue",
			__FILE__, __LINE__);
		goto err;
	}
	BIO_free(outbio);
	if (debug)
		BIO_printf(bio_err, "%s:%d: pending request written to %s\n",
			__FILE__, __LINE__, filename);

	/* handle automatic enrollment					*/
	if (scep->automatic) {
		if (debug)
			BIO_printf(bio_err, "%s:%d: automatic enrollment in "
				"effect\n", __FILE__, __LINE__);

		/* verify the caller					*/
		if (check_challenge(scep) == 0) {
			/* automatically grant the request		*/
			if (debug)
				BIO_printf(bio_err, "%s:%d: automatic grant\n",
					__FILE__, __LINE__);
			if (cert_grant(scep) < 0) {
				BIO_printf(bio_err, "%s:%d: grant failed\n",
					__FILE__, __LINE__);
				msg->failinfo =  SCEP_FAILURE_BADCERTID;
				badreply(scep, SCEP_PKISTATUS_FAILURE);
				return 0;
			}
			goodreply(scep, 1);
			return 0;
		}
	}

	if (debug)
		BIO_printf(bio_err, "%s:%d: automatic enrollment "
			"denied\n", __FILE__, __LINE__);
	syslog(LOG_DEBUG, "%s:%d: automatic enrollment denied",
		__FILE__, __LINE__);

	/* prepare a pending reply					*/
	badreply(scep, SCEP_PKISTATUS_PENDING);
	return 0;

	/* but we may as well fail					*/
err:
	badreply(scep, SCEP_PKISTATUS_FAILURE);
	return 0;
}
