/*
 * scep.c -- scep command line client
 *
 * (c) 2001 Dr. Andreas Mueller, Beratung und Entwicklung
 *
 * $Id: scep.c,v 1.9 2001/03/26 10:36:46 afm Exp $
 */
#include <stdlib.h>
#include <stdio.h>
#include <config.h>
#include <init.h>
#include <openssl/bio.h>
#include <openssl/pem.h>
#include <openssl/x509.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#include <selfsigned.h>
#include <createreq.h>
#include <unistd.h>
#include <fingerprint.h>
#include <http.h>
#include <encode.h>
#include <decode.h>

extern int	optind;
extern char	*optarg;

static X509	*extract_cert(scep_t *scep) {
	PKCS7		*p7;
	STACK_OF(X509)	*certs;
	X509		*x;
	int		i;

	/* get the structure						*/
	p7 = scep->reply.rd.p7;

	/* get the certificate from this pkcs#7				*/
	certs = p7->d.sign->cert;
	for (i = 0; i < sk_X509_num(certs); i++) {
		x = sk_X509_value(certs, i);
		if (X509_NAME_cmp(X509_get_issuer_name(x),
			X509_get_subject_name(scep->cacert))) {
			goto not;
		}
		if (X509_NAME_cmp(X509_get_subject_name(x),
			X509_REQ_get_subject_name(scep->clientreq))) {
			goto not;
		}
		return x;
	not:
		BIO_printf(bio_err, "%s:%d: certificate does not match\n",
			__FILE__, __LINE__);
	}
	return NULL;
}

int	main(int argc, char *argv[]) {
	int		c, poll = 0;
	char		*cacertfile = NULL, *keyfile = NULL, *challenge = NULL,
			*requestfile = NULL, *dn = NULL;
	scep_t		scep;
	BIO		*bio, *repbio;
	char		*url = "http://localhost/cgi-bin";
	scepmsg_t	*msg;
	unsigned char	*checkNonce = NULL;

	/* initialize what you can					*/
	scepinit();
	scep_clear(&scep);

	/* we are a client 						*/
	scep.client = 1;

	/* parse command line						*/
	while (EOF != (c = getopt(argc, argv, "dc:r:k:w:pu:")))
		switch (c) {
		case 'd':
			debug++;
			break;
		case 'c':
			cacertfile = optarg;
			break;
		case 'r':
			/* the request file will also contain the self	*/
			/* signed certificate				*/
			requestfile = optarg;
			break;
		case 'k':
			keyfile = optarg;
			break;
		case 'w':
			challenge = optarg;
			break;
		case 'p':
			poll = 1;
			break;
		case 'u':
			url = optarg;
			break;
		}

	/* we are preparing the request message				*/
	msg = &scep.request;

	/* decode the URL						*/
	if (parseurl(&scep, url) < 0) {
		BIO_printf(bio_err, "%s:%d: cannot parse url\n", __FILE__,
			__LINE__);
		goto err;
	}
	if (debug)
		BIO_printf(bio_err, "%s:%d: decoded URL %s|%d|%s\n", __FILE__,
			__LINE__, scep.h.httphost, scep.h.httpport,
			scep.h.httppath);

	/* read private key (this probably needs a password)		*/
	if (keyfile == NULL) {
		BIO_printf(bio_err, "%s:%d: key file is required argument\n",
			__FILE__, __LINE__);
		goto err;
	}
	bio = BIO_new(BIO_s_file());
	if (BIO_read_filename(bio, keyfile) <= 0) {
		BIO_printf(bio_err, "%s:%d: cannot open private key "
			"file\n", __FILE__, __LINE__);
		goto err;
	}
	scep.clientpkey = PEM_read_bio_PrivateKey(bio, NULL, NULL,
		NULL);
	if (scep.clientpkey == NULL) {
		BIO_printf(bio_err, "%s:%d: no private key available\n",
			__FILE__, __LINE__);
		goto err;
	}
	BIO_free(bio);
	if (debug)
		BIO_printf(bio_err, "%s:%d: private key decoded\n",
			__FILE__, __LINE__);

	/* set the transaction id value					*/
	scep.transId = key_fingerprint(scep.clientpkey);
	if (debug)
		BIO_printf(bio_err, "%s:%d: transaction ID is %s\n",
			__FILE__, __LINE__, scep.transId);


	/* read the CA certificate file					*/
	if (cacertfile != NULL) {
		bio = BIO_new(BIO_s_file());
		if (BIO_read_filename(bio, cacertfile) <= 0) {
			BIO_printf(bio_err, "%s:%d: cannot open CA "
				"certificate file\n", __FILE__, __LINE__);
			goto err;
		}
		scep.cacert = PEM_read_bio_X509(bio, NULL, NULL, NULL);
		if (scep.cacert == NULL) {
			BIO_printf(bio_err, "%s:%d: cannot decode CA "
				"certificate\n", __FILE__, __LINE__);
			goto err;
		}
		BIO_free(bio);
		if (debug)
			BIO_printf(bio_err, "%s:%d: CA certificate decoded\n",
				__FILE__, __LINE__);
	}
	if (poll)
		goto pollinginit;

	/* next argument must be the distinguished name, but only if	*/
	/* this is the initial request					*/
	if ((argc - optind) != 1) {
		BIO_printf(bio_err, "%s:%d: expecting one argument: DN\n",
			__FILE__, __LINE__);
		goto err;
	}
	dn = argv[optind];
	if (debug)
		BIO_printf(bio_err, "%s:%d: distinguished name is %s\n",
			__FILE__, __LINE__, dn);


	/* create a certificate request					*/
	if (createreq(&scep, dn, challenge) < 0) {
		BIO_printf(bio_err, "%s:%d: failed to create request\n",
			__FILE__, __LINE__);
		goto err;
	}
	if (debug)
		BIO_printf(bio_err, "%s:%d: certificate request ready\n",
			__FILE__, __LINE__);

	/* pack the request as a PKSCReq message, of type PKCSReq 	*/
	msg->messageType = SCEP_MESSAGE_TYPE_PKCSREQ;
	msg->rd.req = scep.clientreq;

	/* write the request to the request file, for later perusal	*/
	if (requestfile) {
		BIO	*reqbio;
		reqbio = BIO_new(BIO_s_file());
		BIO_write_filename(reqbio, requestfile);
		PEM_write_bio_X509_REQ(reqbio, scep.clientreq);
		BIO_free(reqbio);
	}

	goto common;

pollinginit:
	/* when polling we should read the request from request file	*/
	/* (only needed for the distinguished name of the client)	*/
	if (requestfile) {
		BIO	*reqbio;
		reqbio = BIO_new(BIO_s_file());
		BIO_read_filename(reqbio, requestfile);
		scep.clientreq = PEM_read_bio_X509_REQ(reqbio, NULL,
			NULL, NULL);
	} else {
		BIO_printf(bio_err, "%s:%d: need a request file when polling\n",
			__FILE__, __LINE__);
		goto err;
	}
	if (debug)
		BIO_printf(bio_err, "%s:%d: got a passivated request\n",
			__FILE__, __LINE__);

	/* when polling, the request is a GetCertInitial message	*/
	msg->messageType = SCEP_MESSAGE_TYPE_GETCERTINITIAL;

	/* the contents is the pair issuer and subject			*/
	msg->rd.is = (issuer_and_subject_t *)malloc(
		sizeof(issuer_and_subject_t));
	msg->rd.is->issuer = X509_get_subject_name(scep.cacert);
	msg->rd.is->subject = X509_REQ_get_subject_name(scep.clientreq);

common:
	/* create a self signed certificate for use with SCEP		*/
	if (selfsigned(&scep) < 0) {
		BIO_printf(bio_err, "%s:%d: failed to create self signed "
			"certificate\n", __FILE__, __LINE__);
	}
	if (debug)
		BIO_printf(bio_err, "%s:%d: self signed certificate created\n",
			__FILE__, __LINE__);

	/* set the senderNonce						*/
	scep.senderNonceLength = 16;
	scep.senderNonce = (unsigned char *)malloc(scep.senderNonceLength);
	RAND_bytes(scep.senderNonce, scep.senderNonceLength);
	if (debug)
		BIO_printf(bio_err, "%s:%d: senderNonce set\n", __FILE__,
			__LINE__);
	checkNonce = scep.senderNonce;

	/* all messages sent from the client are base 64 encoded	*/
	msg->base64 = 1;

	/* encode							*/
	if (encode(&scep) < 0) {
		BIO_printf(bio_err, "%s:%d: encoding the request failed\n",
			__FILE__, __LINE__);
		goto err;
	}
	if (debug)
		BIO_printf(bio_err, "%s:%d: encoded bytes: %d\n", 
			__FILE__, __LINE__, scep.request.length);

	/* send the request to the server, read the reply		*/
	repbio = getrequest(&scep);
	if (repbio == NULL) {
		BIO_printf(bio_err, "%s:%d: failed to read correct reply\n",
			__FILE__, __LINE__);
		goto err;
	}

	/* analyze  the reply						*/
	if (decode(&scep, repbio) < 0) {
		BIO_printf(bio_err, "%s:%d: decoding the reply failed\n",
			__FILE__, __LINE__);
		goto err;
	}

	/* display some information about the reply			*/
	printf("transaction id: %s\n", scep.transId);
	printf("PKIstatus: %s\n", scep.reply.pkiStatus);
	printf("reply message type: %s\n", scep.reply.messageType);
	if (scep.reply.failinfo) {
		printf("failinfo: %s\n", scep.reply.failinfo);
	}

	/* make sure we get a CertRep message back			*/
	if (strcmp(scep.reply.messageType, SCEP_MESSAGE_TYPE_CERTREP)) {
		BIO_printf(bio_err, "%s:%d: only CertRep message acceptable "
			" in response to PKCSReq/GetCertInitial\n",
			__FILE__, __LINE__);
		goto err;
	}

	/* check for the Nonces						*/
	if (memcmp(checkNonce, scep.recipientNonce, 16)) {
		BIO_printf(bio_err, "%s:%d: recipientNonce != sent "
			"senderNonce\n", __FILE__, __LINE__);
		goto err;
	}
	if (debug)
		BIO_printf(bio_err, "%s:%d: Nonce check OK\n", __FILE__, 
			__LINE__);

	switch (atoi(scep.reply.pkiStatus)) {
	case 0:
		if (debug)
			BIO_printf(bio_err, "%s:%d: certificate returned\n",
				__FILE__, __LINE__);
		scep.clientcert = extract_cert(&scep);
		if (scep.clientcert) {
			BIO	*cb;
			cb = BIO_new(BIO_s_file());
			BIO_set_fp(cb, stdout, BIO_NOCLOSE);
			PEM_write_bio_X509(cb, scep.clientcert);
			BIO_free(cb);
		}
		exit(EXIT_SUCCESS);
		break;
	case 2:
		if (debug)
			BIO_printf(bio_err, "%s:%d: request failed: %s\n",
				__FILE__, __LINE__, scep.reply.failinfo);
		exit(1);
		break;
	case 3:
		if (debug)
			BIO_printf(bio_err, "%s:%d: request still pending\n",
				__FILE__, __LINE__);
		exit(2);
		break;
	}

	/* error return							*/
err:
	ERR_print_errors(bio_err);
	exit(EXIT_FAILURE);
}
